/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include "gdbus.h"

GDbus::GDbus()
{
	this->m_pConnection = NULL;
	this->m_pCancellable = NULL;
}

GDbus::~GDbus()
{
	GDBusConnection *conn = this->m_pConnection;
	GCancellable *cancel = this->m_pCancellable;

	if (cancel) {
		g_cancellable_cancel(cancel);
		g_object_unref(cancel);
		cancel = NULL;
	}

	if (conn) {
		g_object_unref(conn);
		conn = NULL;
	}
}

error_e GDbus::Create(void)
{
	GError *err = NULL;

#if !GLIB_CHECK_VERSION(2, 36, 0)
	g_type_init();
#endif

	this->m_pConnection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
	if (this->m_pConnection == NULL) {
		if (err != NULL) {
			GLOGD("Failed to connect to the D-BUS daemon [%s]", err->message);
			g_error_free(err);
		}

		return ERROR_OPERATION_FAILED;
	}

	this->m_pCancellable = g_cancellable_new();

	return ERROR_NONE;
}

error_e GDbus::Destroy(void)
{
	g_cancellable_cancel(this->m_pCancellable);
	g_object_unref(this->m_pCancellable);
	this->m_pCancellable = NULL;

	g_object_unref(this->m_pConnection);
	this->m_pConnection = NULL;

	return ERROR_NONE;
}

GDBusConnection *GDbus::GetConnection(void)
{
	return this->m_pConnection;
}

GCancellable *GDbus::GetCancellable(void)
{
	return this->m_pCancellable;
}

error_e GDbus::ConvertErrorStringToEnum(const char *error)
{
	if (NULL != strstr(error, "NoReply"))
		return ERROR_INVALID_OPERATION;
	else if (NULL != strstr(error, "Failed"))
		return ERROR_OPERATION_FAILED;
	else if (NULL != strstr(error, "UnknownMethod"))
		return ERROR_INVALID_OPERATION;
	else if (NULL != strstr(error, "InvalidArguments"))
		return ERROR_INVALID_PARAMETER;
	else if (NULL != strstr(error, "AccessDenied"))
		return ERROR_PERMISSION_DENIED;
	else if (NULL != strstr(error, "PermissionDenied"))
		return ERROR_PERMISSION_DENIED;
	else if (NULL != strstr(error, "NotSupported"))
		return ERROR_NOT_SUPPORTED;
	else if (NULL != strstr(error, "InProgress"))
		return ERROR_IN_PROGRESS;

	return ERROR_OPERATION_FAILED;
}

GVariant *GDbus::InvokeMethod_with_fd(GUnixFDList *fd_list, const char *dest, const char *path,
	const char *iface_name, const char *method, GVariant *params, error_e *dbus_error)
{
	GError *error = NULL;
	GVariant *reply = NULL;
	GDBusConnection *connection = NULL;
	*dbus_error = ERROR_NONE;

	connection = GetConnection();
	if (connection == NULL) {
		GLOGD("GDBusconnection is NULL");
		*dbus_error = ERROR_NOT_INITIALIZED;
		return reply;
	}

	reply = g_dbus_connection_call_with_unix_fd_list_sync(connection,
			dest,
			path,
			iface_name,
			method,
			params,
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_REPLY_TIMEOUT,
			fd_list,
			NULL,
			GetCancellable(),
			&error);

	if (reply == NULL) {
		if (error != NULL) {
			GLOGD("g_dbus_connection_call_with_unix_fd_list_sync() failed "
				"error [%d: %s]", error->code, error->message);
			*dbus_error = ConvertErrorStringToEnum(error->message);
			g_error_free(error);
		} else {
			GLOGD("g_dbus_connection_call_with_unix_fd_list_sync() failed");
			*dbus_error = ERROR_OPERATION_FAILED;
		}

		return NULL;
	}

	return reply;
}

GVariant *GDbus::InvokeMethod(const char *dest, const char *path,
	const char *iface_name, const char *method, GVariant *params, error_e *dbus_error)
{
	GError *error = NULL;
	GVariant *reply = NULL;
	GDBusConnection *connection = NULL;
	*dbus_error = ERROR_NONE;

	connection = GetConnection();
	if (connection == NULL) {
		GLOGD("GDBusconnection is NULL");
		*dbus_error = ERROR_NOT_INITIALIZED;
		return reply;
	}

	reply = g_dbus_connection_call_sync(connection,
			dest,
			path,
			iface_name,
			method,
			params,
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_REPLY_TIMEOUT,
			GetCancellable(),
			&error);

	if (reply == NULL) {
		if (error != NULL) {
			GLOGD("g_dbus_connection_call_sync() failed "
				"error [%d: %s]", error->code, error->message);
			*dbus_error = ConvertErrorStringToEnum(error->message);
			g_error_free(error);
		} else {
			GLOGD("g_dbus_connection_call_sync() failed");
			*dbus_error = ERROR_OPERATION_FAILED;
		}

		return NULL;
	}

	return reply;
}

error_e GDbus::InvokeMethodNonblock(const char *dest, const char *path,
	const char *iface_name, const char *method, GVariant *params, int timeout,
	GAsyncReadyCallback notify_func, void *user_data)
{
	GDBusConnection *connection = NULL;

	connection = GetConnection();
	if (connection == NULL) {
		GLOGD("GDBusconnection is NULL");
		return ERROR_NOT_INITIALIZED;
	}

	g_dbus_connection_call(connection,
			dest,
			path,
			iface_name,
			method,
			params,
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			timeout,
			GetCancellable(),
			(GAsyncReadyCallback) notify_func,
			(gpointer)user_data);

	return ERROR_NONE;
}
