/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdint.h>
#include <fcntl.h>
#include <gio/gio.h>
#include <gio/gunixfdlist.h>

#include "vpn.h"

Vpnsvc::Vpnsvc()
{
	Create();
}

Vpnsvc::~Vpnsvc()
{
	Destroy();
}

error_e Vpnsvc::vpn_init(const char *iface_name, int iface_name_len, int *result,
			int *h_index, char *h_name)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GError *g_error = NULL;
	GUnixFDList *fd_list = NULL;
	int iface_fd = 0;

	fd_list  = g_unix_fd_list_new();
	if (fd_list == NULL) {
		GLOGD("g_unix_fd_list_new() failed!");
		return ERROR_OPERATION_FAILED ;
	}

	if ((iface_fd = open("/dev/net/tun", O_RDWR)) < 0) {
		GLOGD("tun device open fail\n");
		g_object_unref(fd_list);
		return ERROR_OPERATION_FAILED;
	}

	g_unix_fd_list_append(fd_list, iface_fd, &g_error);
	if (g_error != NULL) {
		close(iface_fd);
		g_object_unref(fd_list);
		return ERROR_OPERATION_FAILED;
	}

	message = InvokeMethod_with_fd(fd_list,
		NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_INIT,
		g_variant_new("(su)", iface_name, iface_name_len),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		close(iface_fd);
		g_object_unref(fd_list);
		return error;
	}

	g_variant_get(message, "(iis)", result, h_index, h_name);
	g_variant_unref(message);
	close(iface_fd);
	g_object_unref(fd_list);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_deinit(const char *dev_name, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_DEINIT,
		g_variant_new("(s)", dev_name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_protect(const char *dev_name, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GError *g_error = NULL;
	GUnixFDList *fd_list = NULL;
	int iface_fd = 0;

	fd_list  = g_unix_fd_list_new();
	if (fd_list == NULL) {
		GLOGD("g_unix_fd_list_new() failed!");
		return ERROR_OPERATION_FAILED ;
	}

	if ((iface_fd = open("/dev/net/tun", O_RDWR)) < 0) {
		GLOGD("tun device open fail\n");
		g_object_unref(fd_list);
		return ERROR_OPERATION_FAILED;
	}

	g_unix_fd_list_append(fd_list, iface_fd, &g_error);
	if (g_error != NULL) {
		close(iface_fd);
		g_object_unref(fd_list);
		return ERROR_OPERATION_FAILED;
	}

	message = InvokeMethod_with_fd(fd_list,
		NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_PROTECT,
		g_variant_new("(s)", dev_name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		close(iface_fd);
		g_object_unref(fd_list);
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);
	close(iface_fd);
	g_object_unref(fd_list);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_up(const char *iface_name, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_UP,
		g_variant_new("(s)", iface_name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_down(const char *iface_name, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_DOWN,
		g_variant_new("(s)", iface_name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_block_networks(GVariant *nets_vpn, int nr_nets_vpn,
		GVariant *nets_orig, int nr_nets_orig, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_BLOCK_NETWORKS,
		g_variant_new("(vuvu)", nets_vpn, nr_nets_vpn,
		nets_orig, nr_nets_orig),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_unblock_networks(int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_UNBLOCK_NETWORKS,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_update_settings(int iface_index, const char *local_ip,
		const char *remote_ip, int mtu, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_UPDATE_SETTINGS,
		g_variant_new("(issu)", iface_index, local_ip,
		remote_ip, mtu),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_add_route(const char *iface_name, const char *route, int prefix,
		int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_ADD_ROUTE,
		g_variant_new("(ssi)", iface_name, route, prefix),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_remove_route(const char *iface_name, const char *route, int prefix,
		int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_REMOVE_ROUTE,
		g_variant_new("(ssi)", iface_name, route, prefix),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_add_dns_server(const char* iface_name, const char *dns_server,
		int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_ADD_DNS_SERVER,
		g_variant_new("(ss)", iface_name, dns_server),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Vpnsvc::vpn_remove_dns_server(const char* iface_name, const char *dns_server,
		int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_VPNSVC_PATH,
		NETCONFIG_VPNSVC_INTERFACE,
		VPN_REMOVE_DNS_SERVER,
		g_variant_new("(ss)", iface_name, dns_server),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

