/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <vconf.h>
#include <vconf-keys.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/ioctl.h>

#include "log.h"
#include "util.h"
#include "netdbus.h"
#include "neterror.h"
#include "emulator.h"
#include "wifi-state.h"
#include "wifi-power.h"
#include "network-state.h"
#include "network-dpm.h"
#include "network-monitor.h"
#include "netsupplicant.h"
#include "clatd-handler.h"
#if defined TIZEN_DEBUG_ENABLE
#include "network-dump.h"
#endif

#include "generated-code.h"
/* Define TCP buffer sizes for various networks */
/* ReadMin, ReadInitial, ReadMax */ /* WriteMin, WriteInitial, WriteMax */
#define NET_TCP_BUFFERSIZE_DEFAULT_READ		"4096 87380 704512"
#define NET_TCP_BUFFERSIZE_DEFAULT_WRITE	"4096 16384 110208"
#define NET_TCP_BUFFERSIZE_WIFI_READ		"524288 1048576 2560000"
#define NET_TCP_BUFFERSIZE_WIFI_WRITE		"524288 1048576 2560000"
#define NET_TCP_BUFFERSIZE_LTE_READ		"524288 1048576 2560000"
#define NET_TCP_BUFFERSIZE_LTE_WRITE		"524288 1048576 2560000"
#define NET_TCP_BUFFERSIZE_UMTS_READ		"4094 87380 704512"
#define NET_TCP_BUFFERSIZE_UMTS_WRITE		"4096 16384 110208"
#define NET_TCP_BUFFERSIZE_HSPA_READ		"4092 87380 704512"
#define NET_TCP_BUFFERSIZE_HSPA_WRITE		"4096 16384 262144"
#define NET_TCP_BUFFERSIZE_HSDPA_READ		"4092 87380 704512"
#define NET_TCP_BUFFERSIZE_HSDPA_WRITE		"4096 16384 262144"
#define NET_TCP_BUFFERSIZE_HSUPA_READ		"4092 87380 704512"
#define NET_TCP_BUFFERSIZE_HSUPA_WRITE		"4096 16384 262144"
#define NET_TCP_BUFFERSIZE_HSPAP_READ		"4092 87380 1220608"
#define NET_TCP_BUFFERSIZE_HSPAP_WRITE		"4096 16384 1220608"
#define NET_TCP_BUFFERSIZE_EDGE_READ		"4093 26280 35040"
#define NET_TCP_BUFFERSIZE_EDGE_WRITE		"4096 16384 35040"
#define NET_TCP_BUFFERSIZE_GPRS_READ		"4096 30000 30000"
#define NET_TCP_BUFFERSIZE_GPRS_WRITE		"4096 8760 11680"

#define NET_TCP_BUFFERSIZE_WIFI_RMEM_MAX	"1048576"
#define NET_TCP_BUFFERSIZE_WIFI_WMEM_MAX	"2097152"
#define NET_TCP_BUFFERSIZE_LTE_RMEM_MAX		"5242880"

#define NET_TCP_BUFFERSIZE_WIFID_WMEM_MAX	"2097152"

#define NET_PROC_SYS_NET_IPV4_TCP_RMEM		"/proc/sys/net/ipv4/tcp_rmem"
#define NET_PROC_SYS_NET_IPv4_TCP_WMEM		"/proc/sys/net/ipv4/tcp_wmem"
#define NET_PROC_SYS_NET_CORE_RMEM_MAX		"/proc/sys/net/core/rmem_max"
#define NET_PROC_SYS_NET_CORE_WMEM_MAX		"/proc/sys/net/core/wmem_max"

#define ROUTE_EXEC_PATH						"/sbin/route"

#define TELEPHONY_SERVICE			"com.tcore.ps"
#define TELEPHONY_MASTER_INTERFACE		TELEPHONY_SERVICE ".master"
#define TELEPHONY_MODEM_INTERFACE		TELEPHONY_SERVICE ".modem"
#define TELEPHONY_PROFILE_INTERFACE		TELEPHONY_SERVICE ".context"
#define TELEPHONY_MASTER_PATH			"/"
#define NET_PROFILE_NAME_LEN_MAX 512

typedef struct {
	char			profile_name[NET_PROFILE_NAME_LEN_MAX];
} net_profile_name_t;

static Network *netconfigstate = NULL;
#if defined TIZEN_DEBUG_ENABLE
static Tcpdump *tcpdump_object = NULL;
#endif
static Battery *battery_object = NULL;

struct netconfig_default_connection {
	char *profile;
	char *ifname;
	char *mac_address;
	char *ipaddress;
	char *ipaddress6;
	char *proxy;
	char *essid;
	unsigned int freq;
	gboolean is_metered;
};

static struct netconfig_default_connection
				netconfig_default_connection_info = { NULL, };

gboolean netconfig_iface_network_state_ethernet_cable_state(gint32 *state);

static gboolean __netconfig_is_connected(GVariantIter *array)
{
	gboolean is_connected = FALSE;
	GVariant *variant = NULL;
	gchar *key = NULL;
	const gchar *value = NULL;

	while (g_variant_iter_loop(array, "{sv}", &key, &variant)) {
		if (g_strcmp0(key, "State") != 0)
			continue;

		if (g_variant_is_of_type(variant, G_VARIANT_TYPE_STRING)) {
			value = g_variant_get_string(variant, NULL);
			if (g_strcmp0(value, "ready") == 0 || g_strcmp0(value, "online") == 0)
				is_connected = TRUE;
		}

		g_free(key);
		g_variant_unref(variant);
		break;
	}

	return is_connected;
}

static char *__netconfig_get_default_profile(void)
{
	GVariant *message = NULL;
	GVariantIter *iter;
	GVariantIter *next;
	gchar *default_profile = NULL;
	gchar *object_path;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetServices", NULL);
	if (message == NULL) {
		ERR("Failed to get profiles");
		return NULL;
	}

	g_variant_get(message, "(a(oa{sv}))", &iter);
	while (g_variant_iter_loop(iter, "(oa{sv})", &object_path, &next)) {
		if (object_path == NULL)
			continue;

		if (netconfig_is_cellular_profile(object_path) && !netconfig_is_cellular_internet_profile(object_path))
			continue;

		if (__netconfig_is_connected(next) == TRUE) {
			default_profile = g_strdup(object_path);
			g_free(object_path);
			g_variant_iter_free(next);
			break;
		}
	}
	g_variant_iter_free(iter);
	g_variant_unref(message);

	return default_profile;
}

static int __netconfig_telephony_get_modem_object_path(GSList **modem_path_list)
{
	GVariant *result;
	GVariantIter *iter_modem = NULL;
	GVariantIter *modem_properties = NULL;
	const char *modem_path;

	result = netconfig_invoke_dbus_method(TELEPHONY_SERVICE, TELEPHONY_MASTER_PATH,
			TELEPHONY_MASTER_INTERFACE, "GetModems", NULL);
	if (result == NULL) {
		ERR("Failed to get modem path list");
		return -1;
	}

	g_variant_get(result, "(a{sa{ss}})", &iter_modem);
	while (g_variant_iter_loop(iter_modem, "{sa{ss}}", &modem_path, &modem_properties)) {
		*modem_path_list = g_slist_append(*modem_path_list, g_strdup(modem_path));
		DBG("modem object path: %s",	modem_path);
	}

	g_variant_iter_free(iter_modem);
	g_variant_unref(result);

	return 0;
}

static int __netconfig_telephony_get_profile_list(net_profile_name_t **profile_list,
		int *profile_count)
{
	int ret = 0;
	int count = 0, i = 0;
	const char *str = NULL;
	GVariant *result;
	GVariantIter *iter = NULL;
	GSList *profiles = NULL, *list = NULL;
	net_profile_name_t *plist = NULL;

	GSList *modem_path_list = NULL;
	const char *path = NULL;

	ret = __netconfig_telephony_get_modem_object_path(&modem_path_list);
	if (ret < 0) {
		ERR("Failed to get modems path list");

		g_slist_free_full(modem_path_list, g_free);
		return ret;
	}

	for (list = modem_path_list; list != NULL; list = list->next) {
		path = (const char *)list->data;

		DBG("path: %s", path);
		result = netconfig_invoke_dbus_method(TELEPHONY_SERVICE, path,
				TELEPHONY_MODEM_INTERFACE, "GetProfileList", NULL);
		if (result == NULL) {
			DBG("Failed to get profiles: %s", path);
			continue;
		}

		g_variant_get(result, "(as)", &iter);
		while (g_variant_iter_loop(iter, "s", &str))
			profiles = g_slist_append(profiles, g_strdup(str));

		g_variant_iter_free(iter);
		g_variant_unref(result);
	}

	g_slist_free_full(modem_path_list, g_free);

	count = g_slist_length(profiles);
	if (count > 0) {
		plist = (net_profile_name_t*)malloc(sizeof(net_profile_name_t) * count);
	} else {
		*profile_count = 0;
		goto out;
	}

	if (plist == NULL) {
		ERR("Failed to allocate memory");
		*profile_count = 0;
		ret = -1;
		goto out;
	}

	for (list = profiles, i = 0; list != NULL; list = list->next, i++)
		g_strlcpy(plist[i].profile_name,
				(const char *)list->data, NET_PROFILE_NAME_LEN_MAX);

	*profile_list = plist;
	*profile_count = count;

out:
	g_slist_free_full(profiles, g_free);

	return ret;
}

static int __netconfig_telephony_search_pdp_profile(const char* profile_name, net_profile_name_t* pdp_name)
{
	int ret;
	net_profile_name_t* profile_list = NULL;
	char* prof_name = NULL;
	char* tel_prof_name = NULL;
	char* found_ptr = NULL;
	int profile_count = 0;
	int i;

	/* Get pdp profile list from telephony service */
	ret = __netconfig_telephony_get_profile_list(&profile_list, &profile_count);
	if (ret < 0) {
		ERR("Failed to get profile list from telephony service");
		g_free(profile_list);
		return ret;
	}

	if (profile_list == NULL || profile_count <= 0) {
		ERR("There is no PDP profiles");
		g_free(profile_list);
		return -1;
	}

	/* Find matching profile */
	prof_name = strrchr(profile_name, '/') + 1;
	for (i = 0; i < profile_count; i++) {
		tel_prof_name = strrchr(profile_list[i].profile_name, '/') + 1;
		found_ptr = strstr(prof_name, tel_prof_name);

		if (found_ptr != NULL && g_strcmp0(found_ptr, tel_prof_name) == 0) {
			g_strlcpy(pdp_name->profile_name,
					profile_list[i].profile_name, NET_PROFILE_NAME_LEN_MAX);

			DBG("PDP profile name found in cellular profile: %s", pdp_name->profile_name);
			break;
		}
	}

	if (i >= profile_count) {
		ERR("There is no matching PDP profiles");
		g_free(profile_list);
		return -1;
	}

	g_free(profile_list);

	return ret;
}

static gboolean __netconfig_telephony_get_metered_info(net_profile_name_t* pdp_name)
{
	GVariant *result;
	GVariantIter *iter;
	const gchar *key = NULL;
	const gchar *value = NULL;
	gboolean ret = FALSE;

	if (pdp_name == NULL) {
		ERR("Invalid parameter!");
		return ret;
	}

	result = netconfig_invoke_dbus_method(TELEPHONY_SERVICE, pdp_name->profile_name,
			TELEPHONY_PROFILE_INTERFACE, "GetProfile", NULL);
	if (result == NULL) {
		ERR("_net_invoke_dbus_method failed");
		return ret;
	}

	g_variant_get(result, "(a{ss})", &iter);
	while (g_variant_iter_next(iter, "{ss}", &key, &value)) {
		if (g_strcmp0(key, "is_metered") == 0) {
			if (value == NULL)
				continue;

			if (g_strcmp0(value, "TRUE") == 0)
				ret = TRUE;
		}
	}

	g_variant_iter_free(iter);
	g_variant_unref(result);

	DBG("is_metered = %s", ret ? "TRUE" : "FALSE");

	return ret;
}

static void __netconfig_get_default_connection_info(const char *profile)
{
	GVariant *message = NULL, *variant = NULL, *variant2 = NULL;
	GVariantIter *iter = NULL, *iter1 = NULL,  *service = NULL;
	GVariant *next = NULL;
	gchar *obj_path;
	gchar *key = NULL;
	gchar *key1 = NULL;
	gchar *key2 = NULL;
	gboolean found_profile = 0;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetServices", NULL);
	if (message == NULL) {
		ERR("Failed to get services informations");
		goto done;
	}

	g_variant_get(message, "(a(oa{sv}))", &service);
	if (service == NULL) {
		ERR("Failed to get services iter");
		goto done;
	}

	while (g_variant_iter_loop(service, "(oa{sv})", &obj_path, &iter)) {
		if (g_strcmp0(obj_path, profile) == 0) {
			g_free(obj_path);
			found_profile = 1;
			break;
		}
	}

	if (iter == NULL || found_profile == 0) {
		ERR("Profile %s doesn't exist", profile);
		goto done;
	}

	while (g_variant_iter_loop(iter, "{sv}", &key, &next)) {
		const gchar *value = NULL;
		guint16 freq = 0;
		if (g_strcmp0(key, "Name") == 0 &&
				netconfig_is_wifi_profile(profile) == TRUE) {
			if (g_variant_is_of_type(next, G_VARIANT_TYPE_STRING)) {
				value = g_variant_get_string(next, NULL);

				netconfig_default_connection_info.essid = g_strdup(value);
			}
		} else if (g_strcmp0(key, "Ethernet") == 0) {
			g_variant_get(next, "a{sv}", &iter1);
			if (iter1 == NULL)
				continue;
			while (g_variant_iter_loop(iter1, "{sv}", &key1, &variant)) {
				if (g_strcmp0(key1, "Interface") == 0) {
					value = g_variant_get_string(variant, NULL);
					netconfig_default_connection_info.ifname = g_strdup(value);
				} else if (g_strcmp0(key1, "Address") == 0) {
					value = g_variant_get_string(variant, NULL);
					netconfig_default_connection_info.mac_address = g_strdup(value);
				}
			}
			g_variant_iter_free(iter1);
		} else if (g_strcmp0(key, "IPv4") == 0) {
			g_variant_get(next, "a{sv}", &iter1);
			if (iter1 == NULL)
				continue;
			while (g_variant_iter_loop(iter1, "{sv}", &key1, &variant)) {
				if (g_strcmp0(key1, "Address") == 0) {
					value = g_variant_get_string(variant, NULL);
					netconfig_default_connection_info.ipaddress = g_strdup(value);
				}
			}
			g_variant_iter_free(iter1);
		} else if (g_strcmp0(key, "IPv6") == 0) {
			g_variant_get(next, "a{sv}", &iter1);
			if (iter1 == NULL)
				continue;
			while (g_variant_iter_loop(iter1, "{sv}", &key1, &variant)) {
				if (g_strcmp0(key1, "Address") == 0) {
					value = g_variant_get_string(variant, NULL);
					netconfig_default_connection_info.ipaddress6 = g_strdup(value);
				}
			}
			g_variant_iter_free(iter1);

		} else if (g_strcmp0(key, "Proxy") == 0) {
			g_variant_get(next, "a{sv}", &iter1);
			if (iter1 == NULL)
				continue;
			while (g_variant_iter_loop(iter1, "{sv}", &key2, &variant2)) {
				GVariantIter *iter_sub = NULL;

				if (g_strcmp0(key2, "Servers") == 0) {
					if (!g_variant_is_of_type(variant2, G_VARIANT_TYPE_STRING_ARRAY)) {
						g_free(key2);
						g_variant_unref(variant2);
						break;
					}

					g_variant_get(variant2, "as", &iter_sub);
					g_variant_iter_loop(iter_sub, "s", &value);
					g_variant_iter_free(iter_sub);
					if (value != NULL && (strlen(value) > 0))
						netconfig_default_connection_info.proxy = g_strdup(value);
				} else if (g_strcmp0(key2, "Method") == 0) {
					if (g_variant_is_of_type(variant2, G_VARIANT_TYPE_STRING)) {
						g_free(key2);
						g_variant_unref(variant2);
						break;
					}

					value = g_variant_get_string(variant2, NULL);
					if (g_strcmp0(value, "direct") == 0) {
						g_free(netconfig_default_connection_info.proxy);
						netconfig_default_connection_info.proxy = NULL;

						g_free(key2);
						g_variant_unref(variant2);
						break;
					}
				}
			}
			g_variant_iter_free(iter1);
		} else if (g_strcmp0(key, "Frequency") == 0) {
			if (g_variant_is_of_type(next, G_VARIANT_TYPE_UINT16)) {
				freq = g_variant_get_uint16(next);
				netconfig_default_connection_info.freq = freq;
			}
		}
	}

	if (netconfig_is_cellular_profile(profile) == TRUE) {
		net_profile_name_t pdp_name;
		int ret;

		ret = __netconfig_telephony_search_pdp_profile(profile, &pdp_name);
		if (ret >= 0 && strlen(pdp_name.profile_name) > 0)
			if (__netconfig_telephony_get_metered_info(&pdp_name))
				netconfig_default_connection_info.is_metered = TRUE;
	}

done:
	if (message)
		g_variant_unref(message);

	if (iter)
		g_variant_iter_free(iter);

	if (service)
		g_variant_iter_free(service);

	return;
}

static char *__netconfig_get_preferred_ipv6_address(char *profile)
{
	GVariant *message = NULL, *variant = NULL, *next = NULL;
	GVariantIter *iter = NULL, *sub_iter = NULL, *service = NULL;
	gchar *obj_path;
	gchar *key = NULL;
	gchar *sub_key = NULL;
	gchar *preferred_address6 = NULL;
	gboolean found_profile = 0;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
					CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
					"GetServices", NULL);
	if (message == NULL) {
		ERR("Failed to get service informations");
		goto done;
	}

	g_variant_get(message, "(a(oa{sv}))", &service);
	if (service == NULL) {
		ERR("Failed to get services iter");
		goto done;
	}

	while (g_variant_iter_loop(service, "(oa{sv})", &obj_path, &iter)) {
		if (g_strcmp0(obj_path, profile) == 0) {
			g_free(obj_path);
			found_profile = 1;
			break;
		}
	}

	if (iter == NULL || found_profile == 0) {
		ERR("Profile %s doesn't exist", profile);
			goto done;
	}

	while (g_variant_iter_loop(iter, "{sv}", &key, &next)) {
		const gchar *value = NULL;
		if (g_strcmp0(key, "IPv6") == 0) {
			g_variant_get(next, "a{sv}", &sub_iter);
			if (sub_iter == NULL)
				continue;
			while (g_variant_iter_loop(sub_iter, "{sv}", &sub_key, &variant)) {
				if (g_strcmp0(sub_key, "Address") == 0) {
					value = g_variant_get_string(variant, NULL);
					if (!preferred_address6)
						preferred_address6 = g_strdup(value);
				}
			}
			g_variant_iter_free(sub_iter);
		}
	}

done:
	if (message)
		g_variant_unref(message);

	if (iter)
		g_variant_iter_free(iter);

	if (service)
		g_variant_iter_free(service);

	return preferred_address6;
}

static void __netconfig_adjust_tcp_buffer_size(void)
{
	int fdr = 0, fdw = 0;
	int fdrmax = 0, fdwmax = 0;
	const char *rbuf_size = NULL;
	const char *wbuf_size = NULL;
	const char *rmax_size = NULL;
	const char *wmax_size = NULL;
	const char *profile = netconfig_get_default_profile();

	if (profile == NULL) {
		DBG("There is no default connection");

		rbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_READ;
		wbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_WRITE;
	} else if (netconfig_is_wifi_profile(profile) == TRUE) {
		DBG("Default connection: Wi-Fi");

		rbuf_size = NET_TCP_BUFFERSIZE_WIFI_READ;
		wbuf_size = NET_TCP_BUFFERSIZE_WIFI_WRITE;
		rmax_size = NET_TCP_BUFFERSIZE_WIFI_RMEM_MAX;
		wmax_size = NET_TCP_BUFFERSIZE_WIFI_WMEM_MAX;
	} else if (netconfig_is_cellular_profile(profile) == TRUE) {
		int telephony_svctype = 0, telephony_pstype = 0;

		netconfig_get_telephony_network_type(&telephony_svctype, &telephony_pstype);
		DBG("Default cellular %d, %d", telephony_svctype, telephony_pstype);

		switch (telephony_pstype) {
		case VCONFKEY_TELEPHONY_PSTYPE_HSPA:
			rbuf_size = NET_TCP_BUFFERSIZE_HSPA_READ;
			wbuf_size = NET_TCP_BUFFERSIZE_HSPA_WRITE;
			break;
		case VCONFKEY_TELEPHONY_PSTYPE_HSUPA:
			rbuf_size = NET_TCP_BUFFERSIZE_HSUPA_READ;
			wbuf_size = NET_TCP_BUFFERSIZE_HSDPA_WRITE;
			break;
		case VCONFKEY_TELEPHONY_PSTYPE_HSDPA:
			rbuf_size = NET_TCP_BUFFERSIZE_HSDPA_READ;
			wbuf_size = NET_TCP_BUFFERSIZE_HSDPA_WRITE;
			break;
#if !defined TIZEN_WEARABLE
		case VCONFKEY_TELEPHONY_PSTYPE_HSPAP:
			rbuf_size = NET_TCP_BUFFERSIZE_HSPAP_READ;
			wbuf_size = NET_TCP_BUFFERSIZE_HSPAP_WRITE;
			break;
#endif
		default:
			switch (telephony_svctype) {
			case VCONFKEY_TELEPHONY_SVCTYPE_LTE:
				rbuf_size = NET_TCP_BUFFERSIZE_LTE_READ;
				wbuf_size = NET_TCP_BUFFERSIZE_LTE_WRITE;
				rmax_size = NET_TCP_BUFFERSIZE_LTE_RMEM_MAX;
				break;
			case VCONFKEY_TELEPHONY_SVCTYPE_3G:
				rbuf_size = NET_TCP_BUFFERSIZE_UMTS_READ;
				wbuf_size = NET_TCP_BUFFERSIZE_UMTS_WRITE;
				break;
			case VCONFKEY_TELEPHONY_SVCTYPE_2_5G_EDGE:
				rbuf_size = NET_TCP_BUFFERSIZE_EDGE_READ;
				wbuf_size = NET_TCP_BUFFERSIZE_EDGE_WRITE;
				break;
			case VCONFKEY_TELEPHONY_SVCTYPE_2_5G:
				rbuf_size = NET_TCP_BUFFERSIZE_GPRS_READ;
				wbuf_size = NET_TCP_BUFFERSIZE_GPRS_WRITE;
				break;
			default:
				/* TODO: Check LTE support */
				rbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_READ;
				wbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_WRITE;
				break;
			}
			break;
		}
	} else {
		DBG("Default TCP buffer configured");

		rbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_READ;
		wbuf_size = NET_TCP_BUFFERSIZE_DEFAULT_WRITE;
	}

	if (rbuf_size != NULL) {
		fdr = open(NET_PROC_SYS_NET_IPV4_TCP_RMEM, O_RDWR | O_CLOEXEC);

		if (fdr < 0 || write(fdr, rbuf_size, strlen(rbuf_size)) < 0)
			ERR("Failed to set TCP read buffer size");

		if (fdr >= 0)
			close(fdr);
	}

	if (wbuf_size != NULL) {
		fdw = open(NET_PROC_SYS_NET_IPv4_TCP_WMEM, O_RDWR | O_CLOEXEC);

		if (fdw < 0 || write(fdw, wbuf_size, strlen(wbuf_size)) < 0)
			ERR("Failed to set TCP write buffer size");

		if (fdw >= 0)
			close(fdw);
	}

	/* As default */
	if (rmax_size == NULL)
		rmax_size = NET_TCP_BUFFERSIZE_WIFI_RMEM_MAX;
	if (wmax_size == NULL)
		wmax_size = NET_TCP_BUFFERSIZE_WIFI_WMEM_MAX;

	if (rmax_size != NULL) {
		fdrmax = open(NET_PROC_SYS_NET_CORE_RMEM_MAX, O_RDWR | O_CLOEXEC);

		if (fdrmax < 0 || write(fdrmax, rmax_size, strlen(rmax_size)) < 0)
			ERR("Failed to set TCP rmem_max size");

		if (fdrmax >= 0)
			close(fdrmax);
	}

	if (wmax_size != NULL) {
		fdwmax = open(NET_PROC_SYS_NET_CORE_WMEM_MAX, O_RDWR | O_CLOEXEC);

		if (fdwmax < 0 || write(fdwmax, wmax_size, strlen(wmax_size)) < 0)
			ERR("Failed to set TCP wmem_max size");

		if (fdwmax >= 0)
			close(fdwmax);
	}
}

static void __netconfig_update_default_connection_info(void)
{
	int old_network_status = 0;
	const char *profile = netconfig_get_default_profile();
	const char *ip_addr = netconfig_get_default_ipaddress();
	const char *ip_addr6 = netconfig_get_default_ipaddress6();
	const char *proxy_addr = netconfig_get_default_proxy();
	unsigned int freq = netconfig_get_default_frequency();
	GVariantBuilder *builder;
	GVariant *params;

	if (emulator_is_emulated() == TRUE) {
		builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

		if (ip_addr != NULL) {
			netconfig_set_vconf_str(VCONFKEY_NETWORK_IP, ip_addr);
			g_variant_builder_add(builder, "{sv}", "IPv4Address",
								  g_variant_new_string(ip_addr));
		} else {
			netconfig_set_vconf_str(VCONFKEY_NETWORK_IP, "");
			g_variant_builder_add(builder, "{sv}", "IPv4Address",
								  g_variant_new_string(""));
		}

		if (ip_addr6 != NULL) {
			netconfig_set_vconf_str(VCONFKEY_NETWORK_IP6, ip_addr6);
			g_variant_builder_add(builder, "{sv}", "IPv6Address",
								  g_variant_new_string(ip_addr6));
		} else {
			netconfig_set_vconf_str(VCONFKEY_NETWORK_IP6, "");
			g_variant_builder_add(builder, "{sv}", "IPv6Address",
								  g_variant_new_string(""));
		}

		params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
		g_variant_builder_unref(builder);

		netconfig_dbus_emit_signal(NULL, NETCONFIG_NETWORK_PATH,
						   NETCONFIG_NETWORK_INTERFACE, "NetworkConfigChanged",
						   params);

		return;
	}

	if (profile == NULL)
		DBG("Reset network state configuration");
	else
		DBG("profile[%s] ipv4(%s) ipv6(%s) proxy(%s)", profile, ip_addr,
			ip_addr6, proxy_addr);

	netconfig_vconf_get_int(VCONFKEY_NETWORK_STATUS, &old_network_status);

	if (profile == NULL && old_network_status != VCONFKEY_NETWORK_OFF) {
		builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

		netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS, VCONFKEY_NETWORK_OFF);
		g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_OFF));

		netconfig_set_vconf_str(VCONFKEY_NETWORK_IP, "");
		g_variant_builder_add(builder, "{sv}", "IPv4Address",
							  g_variant_new_string(""));

		netconfig_set_vconf_str(VCONFKEY_NETWORK_IP6, "");
		g_variant_builder_add(builder, "{sv}", "IPv6Address",
							  g_variant_new_string(""));

		netconfig_set_vconf_str(VCONFKEY_NETWORK_PROXY, "");
		g_variant_builder_add(builder, "{sv}", "ProxyAddress",
							  g_variant_new_string(""));

		params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));

		netconfig_set_vconf_int(VCONFKEY_NETWORK_CONFIGURATION_CHANGE_IND, 0);
		netconfig_dbus_emit_signal(NULL, NETCONFIG_NETWORK_PATH,
						   NETCONFIG_NETWORK_INTERFACE, "NetworkConfigChanged",
						   params);
		netconfig_set_vconf_int("memory/private/wifi/frequency", 0);

		g_variant_builder_unref(builder);

		DBG("Successfully clear IP and PROXY up");

		/* Try to disable CLATD if it was enabled */
		DBG("Disable clatd");
		netconfig_clatd_disable();
	} else if (profile != NULL) {
		char *old_ip = vconf_get_str(VCONFKEY_NETWORK_IP);
		char *old_ip6 = vconf_get_str(VCONFKEY_NETWORK_IP6);
		char *old_proxy = vconf_get_str(VCONFKEY_NETWORK_PROXY);

		builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

		if (netconfig_is_wifi_profile(profile) == TRUE) {
			netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS,
									VCONFKEY_NETWORK_WIFI);
			g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_WIFI));
			netconfig_set_vconf_int("memory/private/wifi/frequency", freq);

			netconfig_set_system_event(SYS_EVT_NETWORK_STATUS,
				EKEY_NETWORK_STATUS, EVAL_NETWORK_WIFI);
		} else if (netconfig_is_cellular_profile(profile)) {

			if (!netconfig_is_cellular_internet_profile(profile)) {
				DBG("connection is not a internet profile - stop to update the cellular state");
				if (old_ip)
					free(old_ip);
				if (old_ip6)
					free(old_ip6);
				if (old_proxy)
					free(old_proxy);
				return;
			}

			netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS,
									VCONFKEY_NETWORK_CELLULAR);
			g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_CELLULAR));

			netconfig_set_system_event(SYS_EVT_NETWORK_STATUS,
				EKEY_NETWORK_STATUS, EVAL_NETWORK_CELLULAR);
			/* Enable clatd if IPv6 is set and no IPv4 address */
			if (!ip_addr && ip_addr6) {
				DBG("Enable clatd");
				netconfig_clatd_enable();
			}
		} else if (netconfig_is_ethernet_profile(profile) == TRUE) {
			netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS,
									VCONFKEY_NETWORK_ETHERNET);
			g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_ETHERNET));
			netconfig_set_system_event(SYS_EVT_NETWORK_STATUS,
				EKEY_NETWORK_STATUS, EVAL_NETWORK_ETHERNET);
		} else if (netconfig_is_bluetooth_profile(profile) == TRUE) {
			netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS,
									VCONFKEY_NETWORK_BLUETOOTH);
			g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_BLUETOOTH));
			netconfig_set_system_event(SYS_EVT_NETWORK_STATUS,
				EKEY_NETWORK_STATUS, EVAL_NETWORK_BT);
		} else{
			netconfig_set_vconf_int(VCONFKEY_NETWORK_STATUS,
									VCONFKEY_NETWORK_OFF);
			g_variant_builder_add(builder, "{sv}", "NetworkStatus",
							  g_variant_new_int32(VCONFKEY_NETWORK_OFF));
			netconfig_set_system_event(SYS_EVT_NETWORK_STATUS,
				EKEY_NETWORK_STATUS, EVAL_NETWORK_DISCONNECTED);
		}

		if (g_strcmp0(old_ip, ip_addr) != 0 || old_ip == NULL) {
			if (ip_addr != NULL) {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_IP, ip_addr);
				g_variant_builder_add(builder, "{sv}", "IPv4Address",
								  g_variant_new_string(ip_addr));
			} else if (old_ip != NULL && strlen(old_ip) > 0) {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_IP, "");
				g_variant_builder_add(builder, "{sv}", "IPv4Address",
								  g_variant_new_string(""));
			}
		}
		if (old_ip)
			free(old_ip);

		if (g_strcmp0(old_ip6, ip_addr6) != 0 || old_ip6 == NULL) {
			if (ip_addr6 != NULL) {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_IP6, ip_addr6);
				g_variant_builder_add(builder, "{sv}", "IPv6Address",
								  g_variant_new_string(ip_addr6));
			} else if (old_ip6 != NULL && strlen(old_ip6) > 0) {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_IP6, "");
				g_variant_builder_add(builder, "{sv}", "IPv6Address",
								  g_variant_new_string(""));
			}
		}
		if (old_ip6)
			free(old_ip6);

		if (g_strcmp0(old_proxy, proxy_addr) != 0) {
			if (proxy_addr == NULL) {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_PROXY, "");
				g_variant_builder_add(builder, "{sv}", "ProxyAddress",
								  g_variant_new_string(""));
			} else {
				netconfig_set_vconf_str(VCONFKEY_NETWORK_PROXY, proxy_addr);
				g_variant_builder_add(builder, "{sv}", "ProxyAddress",
								  g_variant_new_string(proxy_addr));
			}
		}
		if (old_proxy)
			free(old_proxy);

		params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));

		netconfig_set_vconf_int(VCONFKEY_NETWORK_CONFIGURATION_CHANGE_IND, 1);
		netconfig_dbus_emit_signal(NULL, NETCONFIG_NETWORK_PATH,
						   NETCONFIG_NETWORK_INTERFACE, "NetworkConfigChanged",
						   params);

		g_variant_builder_unref(builder);

		DBG("Successfully update default network configuration");
	}

	__netconfig_adjust_tcp_buffer_size();
}

static gboolean __netconfig_is_tech_state_connected(void)
{
	gboolean ret = FALSE;
	GVariant *message = NULL, *variant;
	GVariantIter *iter, *next;
	gchar *path;
	gchar *key;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetTechnologies", NULL);

	if (message == NULL) {
		DBG("Fail to get technology state");
		return FALSE;
	}

	g_variant_get(message, "(a(oa{sv}))", &iter);
	while (g_variant_iter_loop(iter, "(oa{sv})", &path, &next)) {
		if (path == NULL)
			continue;

		while (g_variant_iter_loop(next, "{sv}", &key, &variant)) {
			gboolean data;
			if (g_strcmp0(key, "Connected") == 0) {
				data = g_variant_get_boolean(variant);
				DBG("%s [%s: %s]", path, key, data ? "True" : "False");
				if (TRUE == data) {
					ret = TRUE;
					g_free(path);
					g_free(key);
					g_variant_unref(variant);
					g_variant_iter_free(next);
					goto done;
				}
			}
		}
	}

done:
	g_variant_iter_free(iter);
	g_variant_unref(message);

	return ret;
}

static void __netconfig_update_if_service_connected(void)
{
	GVariant *message = NULL, *var;
	GVariantIter *iter, *next;
	gchar *path;
	gchar *key;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetServices", NULL);

	if (message == NULL) {
		ERR("Failed to get services");
		return;
	}

	g_variant_get(message, "(a(oa{sv}))", &iter);
	while (g_variant_iter_loop(iter, "(oa{sv})", &path, &next)) {
		if (path == NULL)
			continue;

		if (g_str_has_prefix(path,
						CONNMAN_WIFI_SERVICE_PROFILE_PREFIX) == TRUE) {
			if (g_strrstr(path + strlen(CONNMAN_WIFI_SERVICE_PROFILE_PREFIX),
							"hidden") != NULL) {
				/* skip hidden profiles */
				continue;
			}
			/* Process this */
		} else if (g_str_has_prefix(path,
						CONNMAN_CELLULAR_SERVICE_PROFILE_PREFIX) == TRUE) {
			/* Process this */
		} else if (g_str_has_prefix(path,
						CONNMAN_ETHERNET_SERVICE_PROFILE_PREFIX) == TRUE) {
			/* Process this */
		} else {
			continue;
		}

		while (g_variant_iter_loop(next, "{sv}", &key, &var)) {
			if (g_strcmp0(key, "State") == 0) {
				const gchar *sdata = NULL;
				sdata = g_variant_get_string(var, NULL);
				DBG("%s [%s: %s]", path, key, sdata);

				if (g_strcmp0(sdata, "online") == 0 || g_strcmp0(sdata, "ready") == 0) {

					/* Found a connected WiFi / 3G service.
					 * Lets update the default profile info.
					 */
					netconfig_update_default_profile((const gchar*)path);
					g_free(key);
					g_free(path);
					g_variant_unref(var);
					g_variant_iter_free(next);
					goto done;
				}
			}
		}
	}
done:
	g_variant_iter_free(iter);
	g_variant_unref(message);

	return;
}

static void __netconfig_network_notify_result(const char *sig_name, const char *key)
{
	GVariantBuilder *builder;
	GVariant *params;

	INFO("[Signal] %s %s", sig_name, key);

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(builder, "{sv}", "key", g_variant_new_string(key));

	params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	netconfig_dbus_emit_signal(NULL,
				NETCONFIG_NETWORK_PATH,
				NETCONFIG_NETWORK_INTERFACE,
				sig_name,
				params);

	return;
}

const char *netconfig_get_default_profile(void)
{
	return netconfig_default_connection_info.profile;
}

const char *netconfig_get_default_ifname(void)
{
	return netconfig_default_connection_info.ifname;
}

const char *netconfig_get_default_ipaddress(void)
{
	return netconfig_default_connection_info.ipaddress;
}

const char *netconfig_get_default_ipaddress6(void)
{
	return netconfig_default_connection_info.ipaddress6;
}

const char *netconfig_get_default_proxy(void)
{
	return netconfig_default_connection_info.proxy;
}

unsigned int netconfig_get_default_frequency(void)
{
	return netconfig_default_connection_info.freq;
}

const char *netconfig_get_default_mac_address(void)
{
	return netconfig_default_connection_info.mac_address;
}

const char *netconfig_wifi_get_connected_essid(const char *default_profile)
{
	if (default_profile == NULL)
		return NULL;

	if (netconfig_is_wifi_profile(default_profile) != TRUE)
		return NULL;

	if (g_strcmp0(default_profile, netconfig_default_connection_info.profile) != 0)
		return NULL;

	return netconfig_default_connection_info.essid;
}

gboolean netconfig_get_default_is_metered(void)
{
	return netconfig_default_connection_info.is_metered;
}

static int __netconfig_reset_ipv4_socket(void)
{
	int ret;
	int fd;
	struct ifreq ifr;
	struct sockaddr_in sai;
	const char *ipaddr = netconfig_get_default_ipaddress();
	DBG("ipaddr-[%s]", ipaddr);

	if (!ipaddr)
		return -1;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
		return -1;

	memset(&sai, 0, sizeof(struct sockaddr_in));
	sai.sin_family = AF_INET;
	sai.sin_port = 0;
	if (!inet_aton(ipaddr, &sai.sin_addr)) {
		DBG("fail to inet_aton()");
		close(fd);
		return -1;
	}

	memset(&ifr, 0, sizeof(struct ifreq));
	memcpy(&ifr.ifr_addr, &sai, sizeof(sai));
	g_strlcpy((char *)ifr.ifr_name, WIFI_IFNAME, IFNAMSIZ);

#ifndef SIOCKILLADDR
#define SIOCKILLADDR    0x8939
#endif

	ret = ioctl(fd, SIOCKILLADDR, &ifr);
	if (ret < 0) {
		DBG("fail to ioctl[SIOCKILLADDR]");
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}

void netconfig_update_default_profile(const char *profile)
{
	static char *old_profile = NULL;

	/* It's automatically updated by signal-handler
	 * DO NOT update manually
	 *
	 * It is going to update default connection information
	 */

	if (netconfig_default_connection_info.profile != NULL) {

		if (netconfig_is_wifi_profile(netconfig_default_connection_info.profile))
			__netconfig_reset_ipv4_socket();

		g_free(old_profile);
		old_profile = strdup(netconfig_default_connection_info.profile);

		g_free(netconfig_default_connection_info.profile);
		netconfig_default_connection_info.profile = NULL;

		g_free(netconfig_default_connection_info.ifname);
		netconfig_default_connection_info.ifname = NULL;

		g_free(netconfig_default_connection_info.ipaddress);
		netconfig_default_connection_info.ipaddress = NULL;

		g_free(netconfig_default_connection_info.ipaddress6);
		netconfig_default_connection_info.ipaddress6 = NULL;

		g_free(netconfig_default_connection_info.proxy);
		netconfig_default_connection_info.proxy = NULL;

		g_free(netconfig_default_connection_info.mac_address);
		netconfig_default_connection_info.mac_address = NULL;

		netconfig_default_connection_info.freq = 0;
		netconfig_default_connection_info.is_metered = FALSE;

		g_free(netconfig_default_connection_info.essid);
		netconfig_default_connection_info.essid = NULL;
	}

	/* default profile is NULL and new connected profile is NULL */
	if (!profile) {
		char *tmp_profile = __netconfig_get_default_profile();

		if (tmp_profile && netconfig_is_cellular_profile(tmp_profile) &&
			!netconfig_is_cellular_internet_profile(tmp_profile)) {
			DBG("not a default cellular profile");
			g_free(tmp_profile);
			tmp_profile = NULL;
		}

		if (!tmp_profile) {
			__netconfig_update_default_connection_info();
			return;
		}

		netconfig_default_connection_info.profile = g_strdup(tmp_profile);
		__netconfig_get_default_connection_info(tmp_profile);
		__netconfig_update_default_connection_info();
		g_free(tmp_profile);
		return;
	}

	netconfig_default_connection_info.profile = g_strdup(profile);
	__netconfig_get_default_connection_info(profile);
	__netconfig_update_default_connection_info();
}

void netconfig_update_default(void)
{
	if (__netconfig_is_tech_state_connected() == TRUE)
		__netconfig_update_if_service_connected();
	else
		__netconfig_adjust_tcp_buffer_size();
}

char *netconfig_get_ifname(const char *profile)
{
	GVariant *message = NULL, *variant;
	GVariantIter *iter, *next, *service;
	gchar *obj_path;
	gchar *key;
	gchar *key1;
	const gchar *value = NULL;
	gchar *ifname = NULL;
	gboolean found_profile = 0;

	if (profile == NULL)
		return NULL;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetServices", NULL);
	if (message == NULL) {
		ERR("Failed to get services informations");
		return NULL;
	}

	g_variant_get(message, "(a(oa{sv}))", &service);
	while (g_variant_iter_loop(service, "(oa{sv})", &obj_path, &iter)) {
		if (g_strcmp0(obj_path, profile) == 0) {
			g_free(obj_path);
			found_profile = 1;
			break;
		}
	}

	if (found_profile == 0) {
		ERR("Profile %s doesn't exist", profile);
		g_variant_iter_free(service);
		g_variant_unref(message);
		return NULL;
	}

	while (g_variant_iter_loop(iter, "{sv}", &key, &next)) {
		if (g_strcmp0(key, "Ethernet") == 0) {
			while (g_variant_iter_loop(next, "{sv}", &key1, &variant)) {
				if (g_strcmp0(key1, "Interface") == 0) {
					value = g_variant_get_string(variant, NULL);
					if (!ifname)
						ifname = g_strdup(value);
					g_free(key1);
					g_variant_unref(variant);
					break;
				}
			}
		}
	}

	g_variant_unref(message);

	g_variant_iter_free(service);
	g_variant_iter_free(iter);

	return ifname;
}

/* Check Ethernet Cable Plug-in /Plug-out Status */
void netconfig_network_notify_ethernet_cable_state(const char *key)
{
	__netconfig_network_notify_result("EthernetCableState", key);
}

static gboolean handle_add_route(
		Network *object,
		GDBusMethodInvocation *context,
		gchar *ip_addr,
		gchar *netmask,
		gchar *interface,  gchar *gateway, gint address_family)
{
	const gchar *path = ROUTE_EXEC_PATH;
	gchar gw_str[64] = {0,};
	if (gateway != NULL && strlen(gateway) > 1)
		g_snprintf(gw_str, 64, "gw %s", gateway);
	gchar *const args[] = { "/sbin/route", "add", "-net", ip_addr, gw_str,
		"netmask", netmask, "dev", interface, NULL };
	gchar *const envs[] = { NULL };
	const gchar* buf = NULL;
	gchar* ch = NULL;
	int prefix_len = 0;
	int pos = 0;

	DBG("ip_addr(%s), netmask(%s), interface(%s), gateway(%s)", ip_addr, netmask, interface, gateway);

	switch (address_family) {
	case AF_INET:
		if (ip_addr == NULL || netmask == NULL || interface == NULL) {
			ERR("Invalid parameter");
			netconfig_error_invalid_parameter(context);
			return TRUE;
		}

		if (netconfig_execute_file(path, args, envs) < 0) {
			DBG("Failed to add a new route");
			netconfig_error_permission_denied(context);
			return TRUE;
		}

		break;
	case AF_INET6:
		if (ip_addr == NULL || interface == NULL || gateway == NULL) {
			ERR("Invalid parameter");
			netconfig_error_invalid_parameter(context);
			return TRUE;
		}

		buf = ip_addr;
		ch = strchr(buf, '/');
		pos = ch - buf + 1;
		if (ch) {
			prefix_len = atoi(ch + 1);
			ip_addr[pos-1] = '\0';
		} else {
			prefix_len = 128;
		}

		if (netconfig_add_route_ipv6(ip_addr, interface, gateway, prefix_len) < 0) {
			DBG("Failed to add a new route");
			netconfig_error_permission_denied(context);
			return TRUE;
		}
		break;
	default:
		DBG("Unknown Address Family");
		netconfig_error_invalid_parameter(context);
		return TRUE;
	}

	DBG("Successfully added a new route");
	network_complete_add_route(object, context, TRUE);
	return TRUE;
}

static gboolean handle_remove_route(
		Network *object,
		GDBusMethodInvocation *context,
		gchar *ip_addr,
		gchar *netmask,
		gchar *interface, gchar *gateway, gint address_family)
{
	const char *path = ROUTE_EXEC_PATH;
	gchar gw_str[64] = {0,};
	if (gateway != NULL && strlen(gateway) > 1)
		g_snprintf(gw_str, 64, "gw %s", gateway);
	gchar *const args[] = { "/sbin/route", "del", "-net", ip_addr, gw_str,
		"netmask", netmask, "dev", interface, NULL };
	char *const envs[] = { NULL };
	const char* buf = NULL;
	char* ch = NULL;
	int prefix_len = 0;
	int pos = 0;

	DBG("ip_addr(%s), netmask(%s), interface(%s), gateway(%s)", ip_addr, netmask, interface, gateway);

	switch (address_family) {
	case AF_INET:
		if (ip_addr == NULL || netmask == NULL || interface == NULL) {
			DBG("Invalid parameter!");
			netconfig_error_invalid_parameter(context);
			return TRUE;
		}
		if (netconfig_execute_file(path, args, envs) < 0) {
			DBG("Failed to remove the route");
			netconfig_error_permission_denied(context);
			return TRUE;
		}
		break;
	case AF_INET6:
		if (ip_addr == NULL || interface == NULL || gateway == NULL) {
			DBG("Invalid parameter!");
			netconfig_error_invalid_parameter(context);
			return TRUE;
		}

		buf = ip_addr;
		ch = strchr(buf, '/');
		pos = ch - buf + 1;
		if (ch) {
			prefix_len = atoi(ch + 1);
			ip_addr[pos-1] = '\0';
		} else {
			prefix_len = 128;
		}

		if (netconfig_del_route_ipv6(ip_addr, interface, gateway, prefix_len) < 0) {
			DBG("Failed to remove the route");
			netconfig_error_permission_denied(context);
			return TRUE;
		}
		break;
	default:
		DBG("Unknown Address Family");
		netconfig_error_invalid_parameter(context);
		return TRUE;
	}

	DBG("Successfully removed the route");
	network_complete_remove_route(object, context, TRUE);
	return TRUE;
}

static gboolean handle_check_get_privilege(Network *object,
		GDBusMethodInvocation *context)
{
	network_complete_check_get_privilege(object, context);
	return TRUE;
}


static gboolean handle_check_profile_privilege(Network *object,
		GDBusMethodInvocation *context)
{
	network_complete_check_profile_privilege(object, context);
	return TRUE;
}

static gboolean handle_check_internet_privilege(Network *object,
		GDBusMethodInvocation *context)
{
	network_complete_check_internet_privilege(object, context);
	return TRUE;
}

gboolean handle_ethernet_cable_state(Network *object,
	GDBusMethodInvocation *context)
{
	int ret = 0;
	int state = 0;

	ret = netconfig_get_ethernet_cable_state(&state);
	if (ret != 0) {
		DBG("Failed to get ethernet cable state");
		netconfig_error_fail_ethernet_cable_state(context);
		return TRUE;
	}

	DBG("Successfully get ethernet cable state[%d]", state);
	network_complete_ethernet_cable_state(object, context, state);
	return TRUE;
}

gboolean handle_get_metered_info(Network *object,
	GDBusMethodInvocation *context)
{
	gboolean state = 0;

	state = netconfig_get_default_is_metered();

	DBG("Default metered state [%s]", state ? "TRUE" : "FALSE");
	network_complete_get_metered_info(object, context, state);
	return TRUE;
}

gboolean handle_preferred_ipv6_address(Network *object,
	GDBusMethodInvocation *context, gchar *profile)
{
	char *address = NULL;

	address = __netconfig_get_preferred_ipv6_address(profile);
	if (address == NULL) {
		DBG("Failed to get preferred IPv6 address");
		netconfig_error_fail_preferred_ipv6_address(context);
		return TRUE;
	}

	DBG("Successfully get preferred IPv6 address[%s]", address);
	network_complete_preferred_ipv6_address(object, context, address);
	return TRUE;
}

gboolean handle_get_battery_dn_list(Battery *object,
	GDBusMethodInvocation *context)
{
	GVariantBuilder *builder = NULL;
	GVariant *ret_params = NULL;
	gchar *params_str = NULL;

	builder = g_variant_builder_new(G_VARIANT_TYPE("aa{sv}"));

	netconfig_battery_get_dn_list(builder);

	ret_params = g_variant_builder_end(builder);
	if (ret_params) {
		params_str = g_variant_print(ret_params, TRUE);
		DBG("DN list [%s]", params_str);
	}

	battery_complete_get_battery_dn_list(object, context, ret_params);
	g_variant_builder_unref(builder);
	return TRUE;
}

gboolean handle_get_battery_wifi_list(Battery *object,
	GDBusMethodInvocation *context)
{
	GVariantBuilder *builder = NULL;
	GVariant *ret_params = NULL;
	gchar *params_str = NULL;

	builder = g_variant_builder_new(G_VARIANT_TYPE("aa{sv}"));

	netconfig_battery_get_wifi_list(builder);

	ret_params = g_variant_builder_end(builder);
	if (ret_params) {
		params_str = g_variant_print(ret_params, TRUE);
		DBG("Wi-Fi list [%s]", params_str);
	}

	battery_complete_get_battery_wifi_list(object, context, ret_params);
	g_variant_builder_unref(builder);
	return TRUE;
}

void state_object_create_and_init(void)
{
	DBG("Creating network state object");
	GDBusInterfaceSkeleton *interface_network = NULL;
#if defined TIZEN_DEBUG_ENABLE
	GDBusInterfaceSkeleton *interface_tcpdump = NULL;
#endif
	GDBusInterfaceSkeleton *interface_battery = NULL;
	GDBusConnection *connection = NULL;
	GDBusObjectManagerServer *server = netdbus_get_state_manager();
	if (server == NULL)
		return;

	connection = netdbus_get_connection();
	g_dbus_object_manager_server_set_connection(server, connection);

	/*Interface netconfig.network*/
	netconfigstate = network_skeleton_new();

	interface_network = G_DBUS_INTERFACE_SKELETON(netconfigstate);
	g_signal_connect(netconfigstate, "handle-add-route",
				G_CALLBACK(handle_add_route), NULL);
	g_signal_connect(netconfigstate, "handle-check-get-privilege",
				G_CALLBACK(handle_check_get_privilege), NULL);
	g_signal_connect(netconfigstate, "handle-check-profile-privilege",
				G_CALLBACK(handle_check_profile_privilege), NULL);
	g_signal_connect(netconfigstate, "handle-check-internet-privilege",
				G_CALLBACK(handle_check_internet_privilege), NULL);
	g_signal_connect(netconfigstate, "handle-ethernet-cable-state",
				G_CALLBACK(handle_ethernet_cable_state), NULL);
	g_signal_connect(netconfigstate, "handle-preferred-ipv6-address",
				G_CALLBACK(handle_preferred_ipv6_address), NULL);
	g_signal_connect(netconfigstate, "handle-remove-route",
				G_CALLBACK(handle_remove_route), NULL);
	g_signal_connect(netconfigstate, "handle-launch-mdns",
				G_CALLBACK(handle_launch_mdns), NULL);
	g_signal_connect(netconfigstate, "handle-device-policy-set-wifi",
				G_CALLBACK(handle_device_policy_set_wifi), NULL);
	g_signal_connect(netconfigstate, "handle-device-policy-get-wifi",
				G_CALLBACK(handle_device_policy_get_wifi), NULL);
	g_signal_connect(netconfigstate, "handle-device-policy-set-wifi-profile",
				G_CALLBACK(handle_device_policy_set_wifi_profile), NULL);
	g_signal_connect(netconfigstate, "handle-device-policy-get-wifi-profile",
				G_CALLBACK(handle_device_policy_get_wifi_profile), NULL);
	g_signal_connect(netconfigstate, "handle-get-metered-info",
				G_CALLBACK(handle_get_metered_info), NULL);

	if (!g_dbus_interface_skeleton_export(interface_network, connection,
			NETCONFIG_NETWORK_STATE_PATH, NULL)) {
		ERR("Export with path failed");
	}

#if defined TIZEN_DEBUG_ENABLE
	/*Interface netconfig.tcpdump*/
	tcpdump_object = tcpdump_skeleton_new();

	interface_tcpdump = G_DBUS_INTERFACE_SKELETON(tcpdump_object);
	g_signal_connect(tcpdump_object, "handle-start-tcpdump",
				G_CALLBACK(handle_start_tcpdump), NULL);
	g_signal_connect(tcpdump_object, "handle-stop-tcpdump",
				G_CALLBACK(handle_stop_tcpdump), NULL);
	g_signal_connect(tcpdump_object, "handle-get-tcpdump-state",
				G_CALLBACK(handle_get_tcpdump_state), NULL);

	if (!g_dbus_interface_skeleton_export(interface_tcpdump, connection,
			NETCONFIG_NETWORK_STATE_PATH, NULL)) {
		ERR("Export with path failed");
	}
#endif

	battery_object = battery_skeleton_new();

	interface_battery = G_DBUS_INTERFACE_SKELETON(battery_object);
	g_signal_connect(battery_object, "handle-get-battery-dn-list",
				G_CALLBACK(handle_get_battery_dn_list), NULL);
	g_signal_connect(battery_object, "handle-get-battery-wifi-list",
				G_CALLBACK(handle_get_battery_wifi_list), NULL);

	if (!g_dbus_interface_skeleton_export(interface_battery, connection,
			NETCONFIG_NETWORK_STATE_PATH, NULL)) {
		ERR("Export with path failed");
	}
}

void state_object_deinit(void)
{
	g_object_unref(netconfigstate);
#if defined TIZEN_DEBUG_ENABLE
	g_object_unref(tcpdump_object);
#endif
}

static gboolean __netconfig_check_ipv6_address(const char *address)
{
	unsigned char buf[sizeof(struct in6_addr)];
	int err;

	if (!address)
		return FALSE;

	err = inet_pton(AF_INET6, address, buf);
	if (err > 0)
		return TRUE;

	return FALSE;
}

char *netconfig_get_connected_cellular_internet_ipv6only_profile(struct clatd_ctrl_hint *hint)
{
	GVariant *message = NULL;
	GVariantIter *iter = NULL;
	GVariantIter *next = NULL;
	gchar *cellular_internet_profile = NULL;
	gchar *object_path = NULL;

	message = netconfig_invoke_dbus_method(CONNMAN_SERVICE,
			CONNMAN_MANAGER_PATH, CONNMAN_MANAGER_INTERFACE,
			"GetServices", NULL);
	if (message == NULL) {
		ERR("Failed to get profiles");
		return NULL;
	}

	g_variant_get(message, "(a(oa{sv}))", &iter);
	while (g_variant_iter_loop(iter, "(oa{sv})", &object_path, &next)) {
		if (object_path == NULL)
			continue;

		if (netconfig_is_cellular_profile(object_path) == FALSE)
			continue;

		if (netconfig_is_cellular_internet_profile(object_path) == FALSE)
			continue;

		if (__netconfig_is_connected(next) == TRUE) {
			DBG("found connected profiles");
			gchar *key = NULL;
			gboolean has_v4 = FALSE, has_v6 = FALSE;

			GVariant *variant = NULL;
			while (g_variant_iter_loop(next, "{sv}", &key, &variant)) {
				if (g_strcmp0(key, "IPv4") == 0) {
					GVariantIter *ipv4_iter = NULL;
					GVariant *ipv4_variant = NULL;
					gchar *ipv4_key = NULL;
					const gchar *ipv4_value = NULL;
					g_variant_get(variant, "a{sv}", &ipv4_iter);
					while (g_variant_iter_loop(ipv4_iter, "{sv}",
						&ipv4_key, &ipv4_variant)) {
						if (g_strcmp0(ipv4_key, "Address") == 0) {
							ipv4_value = g_variant_get_string(ipv4_variant, NULL);
							DBG("ipv4: %s", ipv4_value);
							has_v4 = TRUE;
						}
					}
					g_variant_iter_free (ipv4_iter);
				} else if (g_strcmp0(key, "IPv6") == 0) {
					GVariantIter *ipv6_iter = NULL;
					GVariant *ipv6_variant = NULL;
					gchar *ipv6_key = NULL;
					const gchar *ipv6_value = NULL;
					g_variant_get(variant, "a{sv}", &ipv6_iter);
					while (g_variant_iter_loop(ipv6_iter, "{sv}",
						&ipv6_key, &ipv6_variant)) {
						if (g_strcmp0(ipv6_key, "Address") == 0) {
							ipv6_value = g_variant_get_string(ipv6_variant, NULL);
							DBG("ipv6: %s", ipv6_value);
							has_v6 = TRUE;
						}
					}
					g_variant_iter_free (ipv6_iter);
				} else if (hint && g_strcmp0(key, "Nameservers") == 0) {
					int idx = 0;
					GVariantIter *ns_iter = NULL;
					gchar *nameserver = NULL;
					g_variant_get(variant, "as", &ns_iter);
					while(g_variant_iter_loop(ns_iter, "s", &nameserver)) {
						INFO("nameserver[%d]: %s", idx, nameserver);
						if (idx >= MAX_DNS) {
							INFO("index exceeded %d -- skip", idx);
							continue;
						}
						if (__netconfig_check_ipv6_address(nameserver) == TRUE) {
							hint->nameserver[idx] = g_strdup(nameserver);
							idx++;
						} else {
							INFO("malformed address %s -- skip", nameserver);
						}
					}
				}
			}

			if (!has_v4 && has_v6) {
				DBG("found 'ipv6-only/connected/internet cellular profile'");
				cellular_internet_profile = g_strdup(object_path);
				g_free(object_path);
				g_variant_iter_free(next);
				break;
			}
		}
	}
	g_variant_iter_free(iter);
	g_variant_unref(message);

	return cellular_internet_profile;
}
