/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "log.h"
#include "util.h"
#include "netdbus.h"
#include "neterror.h"
#include "network-state.h"
#include "wifi-eap.h"
#include "wifi-power.h"

static struct wifi_authentication_data *wifi_authdata;

gboolean handle_get_sim_imsi(Wifi *wifi, GDBusMethodInvocation *context)
{
	DBG("Get IMSI");

	g_return_val_if_fail(wifi != NULL, TRUE);

	__netconfig_wifi_get_sim_imsi(wifi, context);

	return TRUE;
}

gboolean handle_req_sim_auth(Wifi *wifi, GDBusMethodInvocation *context, GVariant *rand_data)
{
	gboolean result = TRUE;
	GArray *rand_data_garray;
	GVariantIter *iter;
	gint length;
	guchar *out_auth_data;
	guchar byte;
	int i = 0;

	DBG("Request SIM Authentication");

	g_return_val_if_fail(wifi != NULL, TRUE);

	g_variant_get(rand_data, "ay", &iter);
	length = g_variant_iter_n_children(iter);
	out_auth_data = g_new0(guchar, length);

	while (g_variant_iter_loop(iter, "y", &byte)) {
		*(out_auth_data + i) = byte;
		i++;
	}
	g_variant_iter_free(iter);

	rand_data_garray = g_array_sized_new(FALSE, FALSE, sizeof(guchar), length);
	memcpy(rand_data_garray->data, out_auth_data, length);
	g_free(out_auth_data);
	rand_data_garray->len = length;

	result = __netconfig_wifi_req_sim_auth(rand_data_garray, context, &wifi_authdata);
	g_array_free(rand_data_garray, FALSE);

	if (result)
		wifi_complete_req_sim_auth(wifi, context, result);

	return TRUE;
}

gboolean handle_req_aka_auth(Wifi *wifi, GDBusMethodInvocation *context, GVariant *rand_data, GVariant *autn_data)
{
	netconfig_error_e ret = NETCONFIG_NO_ERROR;
	gboolean result = FALSE;
	GVariantIter *iter;
	gint length;
	guchar *out_auth_data;
	guchar byte;
	int i = 0;
	GArray *rand_data_garray;
	GArray *autn_data_garray;

	DBG("Request AKA Authentication");

	g_return_val_if_fail(wifi != NULL, TRUE);

	g_variant_get(rand_data, "ay", &iter);
	length = g_variant_iter_n_children(iter);
	out_auth_data = g_new0(guchar, length);
	while (g_variant_iter_loop(iter, "y", &byte)) {
		*(out_auth_data + i) = byte;
		i++;
	}
	g_variant_iter_free(iter);

	rand_data_garray = g_array_sized_new(FALSE, FALSE, sizeof(guchar), length);
	memcpy(rand_data_garray->data, out_auth_data, length);
	g_free(out_auth_data);
	rand_data_garray->len = length;

	i = 0;
	g_variant_get(autn_data, "ay", &iter);
	length = g_variant_iter_n_children(iter);
	out_auth_data = g_new0(guchar, length);
	while (g_variant_iter_loop(iter, "y", &byte)) {
		*(out_auth_data + i) = byte;
		i++;
	}
	g_variant_iter_free(iter);

	autn_data_garray = g_array_sized_new(FALSE, FALSE, sizeof(guchar), length);
	memcpy(autn_data_garray->data, out_auth_data, length);
	g_free(out_auth_data);
	autn_data_garray->len = length;

	ret = __netconfig_wifi_req_aka_auth(rand_data_garray, autn_data_garray, context, &wifi_authdata);
	if (ret == NETCONFIG_NO_ERROR) {
		result = TRUE;
		wifi_complete_req_aka_auth(wifi, context, result);
	} else if (ret == NETCONFIG_ERROR_FAILED_REQ_SIM_AUTH_WRONG_PARAM) {
		netconfig_error_dbus_method_return(context,
				NETCONFIG_ERROR_FAILED_REQ_SIM_AUTH_WRONG_PARAM, "FailReqSimAuthWrongParam");
	} else {
		netconfig_error_dbus_method_return(context,
				NETCONFIG_ERROR_FAILED_REQ_SIM_AUTH, "FailReqSimAuth");
	}

	g_array_free(rand_data_garray, FALSE);
	g_array_free(autn_data_garray, FALSE);

	return TRUE;
}

gboolean handle_get_sim_auth(Wifi *wifi, GDBusMethodInvocation *context)
{
	DBG("Get SIM Authdata");

	g_return_val_if_fail(wifi != NULL, TRUE);

	__netconfig_wifi_get_sim_authdata(wifi, context, &wifi_authdata);

	return TRUE;
}

gboolean handle_get_aka_auth(Wifi *wifi, GDBusMethodInvocation *context)
{
	DBG("Get AKA Authdata");

	g_return_val_if_fail(wifi != NULL, TRUE);

	__netconfig_wifi_get_aka_authdata(wifi, context, &wifi_authdata);

	return TRUE;
}

EXPORT_SYM void netconfig_complete_get_sim_imsi(void *wifi, GDBusMethodInvocation *context, char *imsi)
{
	wifi_complete_get_sim_imsi((Wifi*)wifi, context, imsi);
}

EXPORT_SYM void netconfig_complete_get_aka_auth(void *wifi, GDBusMethodInvocation *context, GArray *array)
{
	wifi_complete_get_aka_auth((Wifi*)wifi, context, array->data);
}

EXPORT_SYM void netconfig_complete_get_sim_auth(void *wifi, GDBusMethodInvocation *context, GArray *array)
{
	wifi_complete_get_sim_auth((Wifi*)wifi, context, array->data);
}

EXPORT_SYM void netconfig_wifi_power_on()
{
#if defined TIZEN_WEARABLE
	wifi_power_on_wearable(TRUE);
#else
	wifi_power_on();
#endif
}
