/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/wireless.h>
#include <vconf.h>
#include <vconf-keys.h>

#include "log.h"
#include "util.h"
#include "netdbus.h"
#include "wifi-state.h"
#include "network-state.h"
#include "network-statistics.h"
#include "netsupplicant.h"
#include "wifi-indicator.h"

#define VCONFKEY_WIFI_SNR_MIN	-89

#if !defined TIZEN_WEARABLE
#define WIFI_INDICATOR_INTERVAL	1
#else
#define WIFI_INDICATOR_INTERVAL	10
#endif

#if defined TIZEN_WEARABLE
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL1	(19200 * 1024)
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL2	(2560 * 1024)
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL3	(1536 * 1024)
#else
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL1	(19200 * 1024)
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL2	(7680 * 1024)
#define NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL3	(3840 * 1024)
#endif
#define NETCONFIG_PROCWIRELESS					"/proc/net/wireless"

static int netconfig_wifi_rssi = VCONFKEY_WIFI_SNR_MIN;
static guint32 netconfig_wifi_freq = 0;
static guint netconfig_wifi_indicator_timer = 0;

int netconfig_wifi_get_rssi(void)
{
	return netconfig_wifi_rssi;
}

unsigned int netconfig_wifi_get_freq(void)
{
	return netconfig_wifi_freq;
}

static int __netconfig_wifi_update_and_get_rssi(void)
{
	char *if_path = NULL;
	GVariant *message = NULL;
	GVariant *value = NULL;
	gchar *key;
	GVariantIter *iter;
	GVariant *variant;
	gint32 key_value;
	int rssi_dbm = VCONFKEY_WIFI_SNR_MIN;

	if_path = netconfig_wifi_get_supplicant_interface();
	if (if_path == NULL) {
		ERR("Fail to get wpa_supplicant DBus path");
		return 0;
	}

	message = netconfig_invoke_dbus_method(SUPPLICANT_SERVICE, if_path,
			SUPPLICANT_INTERFACE ".Interface", "SignalPoll", NULL);
	if (message == NULL) {
		ERR("Fail to get SignalPoll from wpa_supplicant");
		g_free(if_path);
		return 0;
	}

	g_variant_get(message, "(v)", &value);

	g_variant_get(value, "a{sv}", &iter);
	while (g_variant_iter_loop(iter, "{sv}", &key, &variant)) {

		if (g_strcmp0(key, "rssi") == 0) {
			key_value = g_variant_get_int32(variant);
			rssi_dbm = (int)key_value;
			DBG("Currently signal dbm value [%d]", rssi_dbm);
		} else if (g_strcmp0(key, "frequency") == 0) {
			netconfig_wifi_freq = g_variant_get_uint32(variant);
			DBG("Currently frequency [%u]", netconfig_wifi_freq);
		}
	}

	g_variant_iter_free(iter);

	g_variant_unref(message);

	g_free(if_path);

	netconfig_wifi_rssi = rssi_dbm;

	return rssi_dbm;
}

static int __netconfig_wifi_convert_dbm_to_level_24(int rssi_dbm)
{
	int rssi_level = 0;

	/* Wi-Fi Signal Strength Display (for 2.4G (dB))
	 *
	 * Excellent :     ~ -63
	 * Good      : -64 ~ -74
	 * Weak      : -75 ~ -82
	 * Very weak : -83 ~ -88
	 * No signal : -89 ~
	 */

	if (rssi_dbm >= -63)
		rssi_level = 4;
	else if (rssi_dbm >= -74)
		rssi_level = 3;
	else if (rssi_dbm >= -82)
		rssi_level = 2;
	else if (rssi_dbm >= -88)
		rssi_level = 1;
	else
		rssi_level = 0;

	return rssi_level;
}

static int __netconfig_wifi_convert_dbm_to_level_50(int rssi_dbm)
{
	int rssi_level = 0;

	/* Wi-Fi Signal Strength Display (for 5G (dB))
	 *
	 * Excellent :     ~ -72
	 * Good      : -73 ~ -78
	 * Weak      : -79 ~ -83
	 * Very weak : -84 ~ -88
	 * No signal : -89 ~
	 */

	if (rssi_dbm >= -72)
		rssi_level = 4;
	else if (rssi_dbm >= -78)
		rssi_level = 3;
	else if (rssi_dbm >= -83)
		rssi_level = 2;
	else if (rssi_dbm >= -88)
		rssi_level = 1;
	else
		rssi_level = 0;

	return rssi_level;
}

int netconfig_wifi_rssi_level(const int rssi_dbm)
{
	int snr_level = 0;
	int freq = 0;
	int ret = 0;

	ret = netconfig_vconf_get_int("memory/private/wifi/frequency", &freq);

	if (!ret && freq > 4900)
		snr_level = __netconfig_wifi_convert_dbm_to_level_50(rssi_dbm);
	else
		snr_level = __netconfig_wifi_convert_dbm_to_level_24(rssi_dbm);

	return snr_level;
}

static void __netconfig_wifi_set_rssi_level(const int snr_level)
{
	static int last_snr_level = 0;

	if (snr_level != last_snr_level) {
		netconfig_set_vconf_int(VCONFKEY_WIFI_STRENGTH, snr_level);
		netconfig_battery_update_wifi_rssi(snr_level);
		last_snr_level = snr_level;
	}
}

static void __netconfig_wifi_data_activity_booster(int level)
{
	gboolean reply = FALSE;
	GVariant *params = NULL;
	int level1 = 1;
	int level2 = 2;
	int level3 = 3;

	int lock = 2000;
	int unlock = 0;

	static int old_level = 0;

	if (level < 0)
		return;

	if (level > 0) {
		/* enable booster */
		switch (level) {
		case 1:
			params = g_variant_new("(ii)", level1, lock);
			break;
		case 2:
			params = g_variant_new("(ii)", level2, lock);
			break;
		case 3:
			params = g_variant_new("(ii)", level3, lock);
			break;
		default:
			ERR("Invalid level");
			return;
		}

		reply = netconfig_invoke_dbus_method_nonblock(
				"org.tizen.system.deviced",
				"/Org/Tizen/System/DeviceD/PmQos",
				"org.tizen.system.deviced.PmQos",
				"WifiThroughput",
				params,
				NULL);
		if (reply != TRUE)
			return;
	}

	/* disable previous booster */
	if (old_level == 0 || old_level == level)
		return;

	switch (old_level) {
	case 1:
		params = g_variant_new("(ii)", level1, unlock);
		break;
	case 2:
		params = g_variant_new("(ii)", level2, unlock);
		break;
	case 3:
		params = g_variant_new("(ii)", level3, unlock);
		break;
	default:
		ERR("Invalid level");
		return;
	}

	reply = netconfig_invoke_dbus_method_nonblock(
			"org.tizen.system.deviced",
			"/Org/Tizen/System/DeviceD/PmQos",
			"org.tizen.system.deviced.PmQos",
			"WifiThroughput",
			params,
			NULL);
	if (reply != TRUE)
		return;

	old_level = level;
}

static void __netconfig_wifi_update_indicator(void)
{
	static int last_transfer_state = 0;
	static guint64 netconfig_wifi_tx_bytes = 0;
	static guint64 netconfig_wifi_rx_bytes = 0;
	static int booster_tic = 0;
	static int old_level = 0;
	int booster_level = 0;
	guint64 tx, rx, tx_diff, rx_diff;
	int transfer_state;

	if (netconfig_wifi_get_bytes_statistics(&tx, &rx) == TRUE) {
		tx_diff = tx - netconfig_wifi_tx_bytes;
		rx_diff = rx - netconfig_wifi_rx_bytes;

		if (tx_diff > 0) {
			if (rx_diff > 0)
				transfer_state = VCONFKEY_WIFI_TRANSFER_STATE_TXRX;
			else
				transfer_state = VCONFKEY_WIFI_TRANSFER_STATE_TX;
		} else {
			if (rx_diff > 0)
				transfer_state = VCONFKEY_WIFI_TRANSFER_STATE_RX;
			else
				transfer_state = VCONFKEY_WIFI_TRANSFER_STATE_NONE;
		}

		if (transfer_state != last_transfer_state) {
			netconfig_set_vconf_int(VCONFKEY_WIFI_TRANSFER_STATE, transfer_state);
			last_transfer_state = transfer_state;
		}

		/* NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER */
		if (tx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL1 ||
			rx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL1)
			booster_level = 1;
		else if (tx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL2 ||
				rx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL2)
			booster_level = 2;
		else if (tx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL3 ||
				rx_diff >= NETCONFIG_WIFI_DATA_ACTIVITY_BOOSTER_LEVEL3)
			booster_level = 3;

		if (old_level == booster_level) {
			if (--booster_tic <= 0) {
				__netconfig_wifi_data_activity_booster(booster_level);

				booster_tic = 2;
			}
		} else {
			__netconfig_wifi_data_activity_booster(booster_level);

			if (booster_level > 0)
				booster_tic = 2;
			else
				booster_tic = 0;
		}

		old_level = booster_level;

		netconfig_wifi_tx_bytes = tx;
		netconfig_wifi_rx_bytes = rx;
	}
}

static gboolean __wifi_indicator_monitor(gpointer data)
{
	int rssi_dbm = 0;
	int snr_level = 0;
	int pm_state = VCONFKEY_PM_STATE_NORMAL;

	if (wifi_state_get_service_state() != NETCONFIG_WIFI_CONNECTED)
		return FALSE;

	/* In case of LCD off, we don't need to update Wi-Fi indicator */
	netconfig_vconf_get_int(VCONFKEY_PM_STATE, &pm_state);
	if (pm_state >= VCONFKEY_PM_STATE_LCDOFF)
		return TRUE;

	rssi_dbm = __netconfig_wifi_update_and_get_rssi();
	/* INFO("%d dbm", rssi_dbm); */
	snr_level = netconfig_wifi_rssi_level(rssi_dbm);
	__netconfig_wifi_set_rssi_level(snr_level);

	__netconfig_wifi_update_indicator();

	return TRUE;
}

void netconfig_wifi_indicator_start(void)
{
	INFO("Start Wi-Fi indicator");

	netconfig_set_vconf_int(VCONFKEY_WIFI_STRENGTH, VCONFKEY_WIFI_STRENGTH_MAX);
	__wifi_indicator_monitor(&netconfig_wifi_indicator_timer);
	netconfig_battery_update_wifi_rssi(VCONFKEY_WIFI_STRENGTH_MAX);
	netconfig_start_timer_seconds(WIFI_INDICATOR_INTERVAL, __wifi_indicator_monitor, NULL, &netconfig_wifi_indicator_timer);
}

void netconfig_wifi_indicator_stop(void)
{
	INFO("Stop Wi-Fi indicator");

	netconfig_stop_timer(&netconfig_wifi_indicator_timer);

	netconfig_wifi_rssi = VCONFKEY_WIFI_SNR_MIN;
	netconfig_wifi_freq = 0;
}
