/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <errno.h>
#include <vconf.h>
#include <glib.h>

#include "log.h"
#include "util.h"
#include "netdbus.h"
#include "neterror.h"
#include "wifi-wps.h"
#include "wifi-power.h"
#include "wifi-state.h"
#include "netsupplicant.h"
#include "wifi-background-scan.h"

struct netconfig_wifi_wps {
	char *pin;
	gboolean pbc;
};

static struct netconfig_wifi_wps wifi_wps;

void netconfig_wifi_notify_wps_credentials(const char *ssid, gsize ssid_len, const char *wps_key)
{
	GVariantBuilder *builder;
	GVariant *params;
	const char *sig_name = "WpsCredentials";
	const char *prop_ssid = "ssid";
	const char *prop_key = "key";
	GVariantBuilder *rawssid_builder = NULL;
	int i;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	rawssid_builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for (i = 0; i < ssid_len; i++)
		g_variant_builder_add(rawssid_builder, "y", ssid[i]);
	g_variant_builder_add(builder, "{sv}", prop_ssid, g_variant_new("ay", rawssid_builder));
	g_variant_builder_unref(rawssid_builder);
	g_variant_builder_add(builder, "{sv}", prop_key, g_variant_new_string(wps_key));

	params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	netconfig_dbus_emit_signal(NULL,
				NETCONFIG_WIFI_PATH,
				NETCONFIG_WIFI_INTERFACE,
				sig_name,
				params);

	INFO("Sent signal (%s)", sig_name);
	return;
}

void netconfig_wifi_notify_wps_completed(const char *ssid, gsize ssid_len)
{
	GVariantBuilder *builder;
	GVariant *params;
	const char *sig_name = "WpsCompleted";
	const char *prop_ssid = "ssid";
	GVariantBuilder *rawssid_builder = NULL;
	int i;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	rawssid_builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for (i = 0; i < ssid_len; i++)
		g_variant_builder_add(rawssid_builder, "y", ssid[i]);
	g_variant_builder_add(builder, "{sv}", prop_ssid, g_variant_new("ay", rawssid_builder));
	g_variant_builder_unref(rawssid_builder);

	params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	netconfig_dbus_emit_signal(NULL,
				NETCONFIG_WIFI_PATH,
				NETCONFIG_WIFI_INTERFACE,
				sig_name,
				params);

	INFO("Sent signal (%s)", sig_name);
	return;
}

void netconfig_wifi_notify_wps_fail_event(int config_error, int error_indication)
{
	GVariantBuilder *builder;
	GVariant *params;
	const char *sig_name = "WpsFailEvent";
	const char *prop_config_error = "config_error";
	const char *prop_error_indication = "error_indication";

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(builder, "{sv}", prop_config_error, g_variant_new_int32(config_error));
	g_variant_builder_add(builder, "{sv}", prop_error_indication, g_variant_new_int32(error_indication));

	params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	netconfig_dbus_emit_signal(NULL,
				NETCONFIG_WIFI_PATH,
				NETCONFIG_WIFI_INTERFACE,
				sig_name,
				params);

	INFO("Sent signal (%s)", sig_name);
	return;
}

static void interface_wps_start_result(GObject *source_object,
			GAsyncResult *res, gpointer user_data)
{
	GVariant *reply;
	GDBusConnection *conn = NULL;
	GError *error = NULL;

	conn = G_DBUS_CONNECTION(source_object);
	reply = g_dbus_connection_call_finish(conn, res, &error);

	if (reply == NULL) {
		if (error != NULL) {
			ERR("Fail to request status [%d: %s]",
					error->code, error->message);
			g_error_free(error);
		} else {
			ERR("Fail torequest status");
		}
	} else {
		DBG("Successfully M/W--->WPAS: Interface.WPS.Start Method");
	}

	g_variant_unref(reply);
	netconfig_gdbus_pending_call_unref();
}

static void __netconfig_wifi_invoke_wps_connect(GObject *source_object,
			GAsyncResult *res, gpointer user_data)
{
	GVariant *message = NULL;
	GVariantBuilder *builder = NULL;
	const char *role = "enrollee", *type, *key;
	char *if_path = NULL;
	gboolean reply = FALSE;

	if_path = netconfig_wifi_get_supplicant_interface();
	if (if_path == NULL) {
		DBG("Fail to get wpa_supplicant DBus path");
		return;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

	key = "Role";
	g_variant_builder_add(builder, "{sv}", key, g_variant_new_string(role));

	key = "Type";

	if (wifi_wps.pbc == TRUE)
		type = "pbc";
	else
		type = "pin";

	g_variant_builder_add(builder, "{sv}", key, g_variant_new_string(type));

	if (wifi_wps.pin != NULL) {
		key = "Pin";
		g_variant_builder_add(builder, "{sv}", key, g_variant_new_string(wifi_wps.pin));
	}
	message = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	DBG("[net-config]: TizenMW-->WPAS: .Interface.WPS.Start");
	reply = netconfig_supplicant_invoke_dbus_method_nonblock(
			SUPPLICANT_SERVICE,
			if_path,
			SUPPLICANT_IFACE_WPS,
			"Start",
			message,
			(GAsyncReadyCallback) interface_wps_start_result);

	g_free(if_path);
	if (reply != TRUE)
		ERR("Fail to Scan");

	return;
}

static gboolean __netconfig_wifi_invoke_wps_process_credentials(char *object_path)
{
	gboolean reply = FALSE;
	GVariant *params = NULL;
	const char *interface = SUPPLICANT_IFACE_WPS;
	const char *key = "ProcessCredentials";
	gboolean credentials = TRUE;

	params = g_variant_new("(ssv)", interface, key, g_variant_new_boolean(credentials));

	INFO("[net-config]: TizenMW-->WPAS: .Set");
	reply = netconfig_invoke_dbus_method_nonblock(SUPPLICANT_SERVICE,
			object_path, DBUS_INTERFACE_PROPERTIES,
			"Set", params, __netconfig_wifi_invoke_wps_connect);

	if (reply != TRUE)
		ERR("M/W--->WPAS: Interface.WPS.Set Method Failed");

	return reply;
}

gboolean netconfig_wifi_wps_connect()
{
	char *if_path = NULL;

	if_path = netconfig_wifi_get_supplicant_interface();
	if (if_path == NULL) {
		DBG("Fail to get wpa_supplicant DBus path");
		 return FALSE;
	}

	if (__netconfig_wifi_invoke_wps_process_credentials(if_path) == TRUE) {
		ERR("Wi-Fi WPS Connect started");
		g_free(if_path);
		return TRUE;
	}

	g_free(if_path);
	return FALSE;
}

static void __interface_wps_cancel_result(GObject *source_object,
			GAsyncResult *res, gpointer user_data)
{
	GVariant *reply;
	GDBusConnection *conn = NULL;
	GError *error = NULL;

	conn = G_DBUS_CONNECTION(source_object);
	reply = g_dbus_connection_call_finish(conn, res, &error);

	if (reply == NULL) {
		if (error != NULL) {
			ERR("Fail to request status [%d: %s]",
					error->code, error->message);
			g_error_free(error);
		} else {
			ERR("Fail torequest status");
		}
	} else {
		DBG("Successfully M/W--->WPAS: Interface.WPS.Cancel Method");
	}

	g_variant_unref(reply);
	netconfig_gdbus_pending_call_unref();
}

static gboolean __netconfig_wifi_invoke_wps_cancel()
{
	gboolean reply = FALSE;
	char *if_path = NULL;

	if_path = netconfig_wifi_get_supplicant_interface();
	if (if_path == NULL) {
		DBG("Fail to get wpa_supplicant DBus path");
		return -ESRCH;
	}

	DBG("M/W--->WPAS: Interface.WPS.Cancel Method");

	reply = netconfig_invoke_dbus_method_nonblock(SUPPLICANT_SERVICE,
			if_path, SUPPLICANT_IFACE_WPS,
			"Cancel", NULL, __interface_wps_cancel_result);

	if (reply != TRUE)
		ERR("M/W--->WPAS: Interface.WPS.Cancel Method Failed");

	g_free(if_path);
	return reply;
}

gboolean netconfig_get_wps_field()
{
	return wifi_wps.pbc;
}

gboolean handle_request_wps_cancel(Wifi *wifi, GDBusMethodInvocation *context)
{
	INFO("Received WPS PBC Cancel Request");
	g_return_val_if_fail(wifi != NULL, TRUE);
	__netconfig_wifi_invoke_wps_cancel();

	wifi_complete_request_wps_cancel(wifi, context);
	return TRUE;
}

gboolean handle_request_wps_connect(Wifi *wifi, GDBusMethodInvocation *context, gchar *param)
{
	INFO("Received WPS PBC/PIN Connection Request");

	g_return_val_if_fail(wifi != NULL, TRUE);

	/* Checking the value of pin if param have a string "PBC"
	 * in that scenario PBC will trigger otherwise PIN Connection */

	if (g_strcmp0(param, "PBC") == 0) {
		wifi_wps.pbc = TRUE;
		wifi_wps.pin = NULL;
	} else {
		wifi_wps.pin = g_strdup(param);
		wifi_wps.pbc = FALSE;
	}

	netconfig_wifi_wps_connect();

	wifi_complete_request_wps_connect(wifi, context);
	return TRUE;
}
